import java.io.*;
import java.util.LinkedList;
import java.util.StringTokenizer;

        /** Provide evaluation functions for the calculator.
         * This class corresponds to the Model in a Model-View
         * pattern. No GUI functionality is defined in this class.
         */
class CalculatorModel {
        /** Calculate the result of evaluating expression.
         *  Return the result as a string.
         */
    public String calculate(String expr) throws Exception {
        final Stack numbers = getNumbers();
        final Stack operators = getOperators();

        // Perform the basic tokenization.
        tokenize(expr);

        // Clear the stack.
        while((operators.size() > 0) &&
                (((Operator)(operators.top())).arithOperator())){
            Operator op = (Operator) operators.top();
            operators.pop();
            op.apply(numbers);
        }

        if(operators.size() != 0){
            String error = "Operator stack is not empty: ";
            while(operators.size() != 0){
                Operator op = (Operator) operators.top();
                operators.pop();
                error += op.getOperator()+" ";
            }
            throw new Exception(error);
        }

        if(numbers.size() == 1){
            Number result = (Number) numbers.top();
            numbers.pop();
            return result.toString();
        }
        else{
            String error = "Result stack is not empty: ";
            while(numbers.size() != 0){
                Number num = (Number) numbers.top();
                numbers.pop();
                error += num+" ";
            }
            throw new Exception(error);
        }
    }

        /** Tokenize the given Java expression.
         */
    protected void tokenize(String expr) throws Exception {
        //  Tokenize on the operators, and return the operators.
        StringTokenizer t = new StringTokenizer(expr,Operator.operators,true);

        while(t.hasMoreTokens()){
            String token = t.nextToken();
            if(Operator.isAnOperator(token)){
                if(Operator.isBinary(token)){
                    pushOperator(new BinaryOperator(token));
                }
                else if(Operator.isParenthesis(token)){
                    pushOperator(new Parenthesis(token));
                }
                else if(Operator.isUnary(token)){
                    pushOperator(new UnaryOperator(token));
                }
                else{
                    throw new Exception("Unknown operator: "+token);
                }
            }
            else{
                if(getLastItem() instanceof Number){
                    throw new Exception("Missing operator between "+
                        getLastItem()+" and "+token);
                }
                pushOperand(token);
            }
        }
    }

        /** Push a parsed operand.
         */
    protected void pushOperand(String val){
        // Work out if it is an integer or not.
        Number num = null;
        try{
            num = Integer.decode(val);
        }
        catch(NumberFormatException e){
            num = new Double(val);
        }
        getNumbers().push(num);
        setLastItem(num);
    }

        /** Push a partial result from evaluating a sub-expression.
         */
    protected void pushOperand(Number n){
        getNumbers().push(n);
    }        

        /** Push an operator onto the operator stack.
         */
    protected void pushOperator(Operator o) throws Exception {
        final Stack numbers = getNumbers();
        final Stack operators = getOperators();

        if(o instanceof Parenthesis){
            if(o.getOperator() == '('){
                operators.push(o);
            }
            else{
                while((operators.size() > 0) && 
                        (((Operator)(operators.top())).getOperator() != '(')){
                    Operator op = (Operator) operators.top();
                    operators.pop();
                    op.apply(numbers);
                }
                if(operators.size() > 0){
                    // Remove the opening parenthesis.
                    operators.pop();
                }
                else{
                    throw new Exception("Missing opening parenthesis.");
                }
            }
        }
        else{
            // See if this operator is really a Unary operator.
            boolean isUnary = false;
            Object last = getLastItem();
            if(last == null){
                isUnary = true;
            }
            else if(last instanceof Parenthesis){
                if(((Parenthesis) last).getOperator() == '('){
                    isUnary = true;
                }
            }
            else if(last instanceof Operator){
                isUnary = true;
            }
            else{
                // It isn't.
            }
            if(isUnary){
                // It must be switched to a UnaryOperator.
                o = new UnaryOperator(o);
            }

            // Apply the top operator if it takes precedence.
            while((operators.size() > 0) && 
                    (((Operator)(operators.top())).takesPrecedence(o))){
                Operator op = (Operator) operators.top();
                operators.pop();
                op.apply(numbers);
            }
            operators.push(o);                
        }
        setLastItem(o);
    }

    protected Object getLastItem(){
        return lastItem;
    }

    protected void setLastItem(Object o){
        lastItem = o;
    }

    protected Stack getNumbers(){
        return numbers;
    }

    protected Stack getOperators(){
        return operators;
    }

    // Where the operands are stored.
    private final Stack numbers = new Stack();
    // Where the operators are stored.
    private final Stack operators = new Stack();
    // The last object pushed - Operator or Number.
    // Used to detect parsing errors.
    private Object lastItem;
}

    /** Encapsulate the functionality of a binary operator.
     */
class BinaryOperator extends Operator {
    BinaryOperator(String s) throws Exception {
        super(s);
        if(!isBinary(s)){
            throw new Exception(s+" is not a binary operator.");
        }
    }

    public Number operate(Number lhs, Number rhs) throws Exception {
        int operator = getOperator();
        if((lhs instanceof Integer) && (rhs instanceof Integer)){
            int result;
            switch(operator){
                case '+':
                    result = lhs.intValue()+rhs.intValue();
                    break;
                case '-':
                    result = lhs.intValue()-rhs.intValue();
                    break;
                case '*':
                    result = lhs.intValue()*rhs.intValue();
                    break;
                case '/':
                    result = lhs.intValue()/rhs.intValue();
                    break;
                case '%':
                    result = lhs.intValue()%rhs.intValue();
                    break;
                default:
                    throw new Exception("Unknown binary operator: "+
                                            getOperator());
            }
            return new Integer(result);
        }
        else{
            double result;
            switch(operator){
                case '+':
                    result = lhs.doubleValue()+rhs.doubleValue();
                    break;
                case '-':
                    result = lhs.doubleValue()-rhs.doubleValue();
                    break;
                case '*':
                    result = lhs.doubleValue()*rhs.doubleValue();
                    break;
                case '/':
                    result = lhs.doubleValue()/rhs.doubleValue();
                    break;
                case '%':
                    result = lhs.doubleValue()%rhs.doubleValue();
                    break;
                default:
                    throw new Exception("Unknown binary operator: "+
                                            getOperator());
            }
            return new Double(result);
        }
    }

    public void apply(Stack numbers) throws Exception {
        if(numbers.size() >= 2){
            Number rhs = (Number) numbers.top();
            numbers.pop();
            Number lhs = (Number) numbers.top();
            numbers.pop();
            numbers.push(operate(lhs,rhs));
        }
        else{
            throw new Exception("Insufficient arguments for: "+
                                getOperator());
        }
    }

    protected int precedence(){
        switch(getOperator()){
            case '+': case '-':
                return 2;
            case '*': case '/': case '%':
                return 3;
            default:
                return super.precedence();
        }
    }
}

    /** Encapsulate the functionality of a unary operator.
     */
class UnaryOperator extends Operator {
    UnaryOperator(String s) throws Exception {
        super(s);
        if(!isUnary(s)){
            throw new Exception(s+" is not a unary operator.");
        }
    }

    UnaryOperator(char op) throws Exception {
        this(String.valueOf(op));
    }

    UnaryOperator(Operator o) throws Exception {
        this(o.getOperator());
    }

    public void apply(Stack numbers) throws Exception {
        if(numbers.size() >= 1){
            Number operand = (Number) numbers.top();
            numbers.pop();
            numbers.push(operate(operand));
        }
        else{
            throw new Exception("Insufficient arguments for: "+
                                getOperator());
        }
    }

    public Number operate(Number operand) throws Exception {
        int operator = getOperator();
        if(operand instanceof Integer){
            int result;
            switch(operator){
                case '+':
                    result = operand.intValue();
                    break;
                case '-':
                    result = -operand.intValue();
                    break;
                default:
                    throw new Exception("Unknown unary operator: "+
                                            getOperator());
            }
            return new Integer(result);
        }
        else{
            double result;
            switch(operator){
                case '+':
                    result = operand.doubleValue();
                    break;
                case '-':
                    result = -operand.doubleValue();
                    break;
                default:
                    throw new Exception("Unknown unary operator: "+
                                            getOperator());
            }
            return new Double(result);
        }
    }

    protected int precedence(){
        switch(getOperator()){
            case '+': case '-':
                return 4;
            default:
                return super.precedence();
        }
    }
}

    /** Encapsulate the functionality of a parentheis operator.
     */
class Parenthesis extends Operator {
    Parenthesis(String s) throws Exception {
        super(s);
        if(!isParenthesis(s)){
            throw new Exception(s+" is not a parenthesis.");
        }
    }
}

    /** Encapsulate the common functionality of operators.
     */
abstract class Operator {
    public static final String operators = "+-*/%()";
    public static final String binaryOperators = "+-*/%";
    public static final String unaryOperators = "+-";
    public static final String parentheses = "()";

    Operator(String op){
        setOperator(op.charAt(0));
    }

    public static boolean isUnary(String s){
        return isType(unaryOperators,s);
    }

    public static boolean isBinary(String s){
        return isType(binaryOperators,s);
    }

    public static boolean isParenthesis(String s){
        return isType(parentheses,s);
    }

    // Check that s is an operator.
    public static boolean isAnOperator(String s){
        return isType(operators,s);
    }

    protected static boolean isType(String oplist,String op){
        return oplist.indexOf(op) >= 0;
    }

    public boolean arithOperator(){
        switch(getOperator()){
            case '+': case '-':
            case '*': case '/': case '%':
                return true;
            default:
                return false;
        }
    }

    public Number operate(Number lhs, Number rhs) throws Exception {
        throw new Exception("Inappropriate arguments to "+getOperator());
    }

    public Number operate(Number operand) throws Exception {
        throw new Exception("Inappropriate argument to "+getOperator());
    }

    public void apply(Stack numbers) throws Exception {
        throw new Exception("Inappropriate application of "+getOperator());
    }

    protected int precedence(){
        return 0;
    }

            // nextOp appears to the right of this.
    public boolean takesPrecedence(Operator nextOp){
        if(precedence() > nextOp.precedence()){
            return true;
        }
        else if(precedence() < nextOp.precedence()){
            return false;
        }
        else{
            // Same. True if left associative.
            return this instanceof BinaryOperator;
        }
    }

    public char getOperator(){
        return (char) operator;
    }

    public String toString(){
        return String.valueOf(getOperator());
    }

    protected void setOperator(char op){
        operator = op;
    }
    private char operator;
}

    /** Define an Object stack. This is used to store either
     *  Operators or Operands.
     */
class Stack {
    public void push(Object o){
        getStack().add(o);
    }

    public int size(){
        return getStack().size();
    }

    public Object top(){
        if(size() > 0){
            return getStack().getLast();
        }
        else{
            return null;
        }
    }

    public void pop(){
        if(size() > 0){
            getStack().removeLast();
        }
    }
        
    protected LinkedList getStack(){
        return stack;
    }

    private final LinkedList stack = new LinkedList();
}
